/* ------------------------------------------------------------------------- */
/*
 *  timefmt.h
 *
 *  Copyright (c) 2004 - 2006 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Wed 05 Apr 2006 21:54:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_TIMEFMT_H
#define CLX_TIMEFMT_H

#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include "lexical_cast.h"
#include "time.h"

#define TIMEFMT_STR "%y/%m/%d %H:%M:%S"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_timefmt
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_timefmt {
	public:
		typedef time time_type;
		typedef time::value_type value_type;
		typedef CharT char_type;
		
		typedef typename std::basic_string<CharT, Traits> string_type;
		typedef typename std::map<CharT, value_type*> time_map;
		typedef typename string_type::const_iterator const_iterator;
		
		static const char_type rep = '%';
		static const char_type eos = '\0';
		static const char_type ws = ' ';
		static const char_type zero = '0';
		static const char_type lpos = '-';
		
		// constructor and destructor
		explicit basic_timefmt(const string_type& fmt = TIMEFMT_STR) :
			t_(), tmap_(), fmt_(fmt) { this->init(); }
		
		explicit basic_timefmt(const time_type& t, const string_type& fmt = TIMEFMT_STR) :
			t_(t), tmap_(), fmt_(fmt) { this->init(); }
		
		virtual ~basic_timefmt() {}
		
		// operator
		virtual void reset() { t_ = time_type(); }
		
		virtual time_type& assign(const string_type& s) {
			string_type tmp(s);
			
			const_iterator ps = tmp.begin();
			const_iterator pf = fmt_.begin();
			
			 while (ps != tmp.end() && pf != fmt_.end()) {
				if (*ps == ws) ps++; // skip white space
				else if (*ps == rep) {
					if (pf + 1 != fmt_.end() && *pf == rep && *(pf + 1) == rep) {
						pf += 2;
						ps++;
					}
					else break;
				}
				else if (*pf == rep) {
					string_type flag;
					while (++pf != fmt_.end() && tmap_.find(*pf) == tmap_.end()) flag += *pf;
					if (pf == fmt_.end()) break; // no match
					char_type c = (pf + 1 != fmt_.end()) ? *(pf + 1) : eos;
					*tmap_[*pf++] = this->get(ps, tmp.end(), c, flag);
					if (ps++ == tmp.end()) break;
					pf++;
				}
				else if (*ps == *pf) { ps++; pf++; }
				else break; // no match
			}
			return t_;
		}
		
		virtual time_type& assign(const string_type& s, const string_type& fmt) {
			fmt_ = fmt;
			return this->assign(s);
		}
		
		virtual string_type str() {
			string_type dest;
			for (const_iterator pf = fmt_.begin(); pf != fmt_.end(); pf++) {
				// if the last character is '%', treat this as same as "%%".
				if (*pf == rep && pf + 1 != fmt_.end() && *(pf + 1) != rep) {
					string_type flag;
					while (++pf != fmt_.end() && tmap_.find(*pf) == tmap_.end()) flag += *pf;
					if (pf == fmt_.end()) break; // uncorrect format
					dest += this->put(*tmap_[*pf], flag);
				}
				else dest += *pf;
			}
			return dest;
		}
		
		virtual string_type str(const string_type& fmt) {
			fmt_ = fmt;
			return this->str();
		}
		
		const time_type& data() const { return t_; }
		
	private:
		typedef typename std::basic_stringstream<CharT, Traits> internal_stream;
		
		time_type t_;
		time_map tmap_;
		string_type fmt_;
		
		void init() {
			tmap_['y'] = &t_.year;
			tmap_['m'] = &t_.mon;
			tmap_['d'] = &t_.day;
			tmap_['H'] = &t_.hour;
			tmap_['M'] = &t_.min;
			tmap_['S'] = &t_.sec;
		}
		
		value_type get(const_iterator& next, const_iterator last, char_type c, const string_type& flag) {
			int n = -1;
			if (!flag.empty() && isdigit(flag.at(0))) n = lexical_cast<int>(flag);
			
			string_type tmp;
			while (next != last && *next != c && n-- != 0) {
				tmp += *next;
				next++;
			}
			if (tmp.empty()) return 0;
			return lexical_cast<value_type>(tmp);
		}
		
		string_type put(value_type x, const string_type& flag) {
			internal_stream ss;
			
			setopt(ss, flag);
			ss << x;
			
			return ss.str();
		}
		
		void setopt(internal_stream& ss, const string_type& flag) {
			bool lp = false;
			for (const_iterator pos = flag.begin(); pos != flag.end(); pos++) {
				switch (*pos) {
				case lpos:
					lp = true;
					ss << std::setfill(ws);
					ss << std::left;
					break;
				case ws:
					ss << std::setfill(*pos);
					break;
				case zero:
					if (!lp) ss << std::setfill(*pos);
					break;
				default:
					if (isdigit(*pos)) {
						setwidth(ss, pos, flag.end());
						return;
					}
					break; /* ignore unknown flags */
				}
			}
		}
		
		void setwidth(internal_stream& ss, const_iterator& cur, const_iterator last) {
			std::string digit;
			for (; cur < last && isdigit(*cur); cur++) digit += *cur;
			if (!digit.empty()) ss << std::setw(lexical_cast<int>(digit));
		}
	};
	
	typedef basic_timefmt<char> timefmt;
}

#endif // CLX_TIMEFMT_H
